@@ -8,7 +8,8 @@ class F1ScoreMetric(tf.keras.metrics.Metric):
                  skip_nan=False, tboard_writer=None, **kwargs):
         super(F1ScoreMetric, self).__init__(name=name, **kwargs)
         self.f1_scores = self.add_weight(name="f1score", initializer='zeros')
-        self.con_mat = tf.Variable(initial_value=np.zeros((n_classes, n_classes)), dtype=tf.float32, name="geometric_f1score_con_mat")
+        self.con_mat = tf.Variable(initial_value=np.zeros((n_classes, n_classes)), dtype=tf.float32,
+                                   name="geometric_f1score_con_mat", aggregation=tf.VariableAggregation.SUM)
         self.n_classes = n_classes
         self.debug = debug
         self.skip_nan = skip_nan
@@ -23,6 +24,7 @@ class F1ScoreMetric(tf.keras.metrics.Metric):
                                                    predictions=y_pred,
                                                    num_classes=self.n_classes),
                           dtype=tf.float32)
+
         self.con_mat.assign_add(con_mat)
 
         tp = tf.linalg.diag_part(self.con_mat)
